using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    class ModelExFrame : Frame
    {
        public ModelExContainer ModelExContainer
        {
            get { return (ModelExContainer)MeshContainer; }
            set { MeshContainer = (MeshContainer)value; }
        }
    }

    class ModelExContainer : MeshContainer
    {
        protected int[] m_aiEffects;
        protected EffectInstance[] m_axEffectInstancesBackup;
        protected int[] m_aiTextures;
        protected Vector3 m_xCentre;
        protected float m_fRadius;
        protected Vector3 m_xScale;
        public int[] Effects
        {
            get { return m_aiEffects; }
            set { m_aiEffects = value; }
        }
        public EffectInstance[] GetEffectInstancesBackup()
        {
            return m_axEffectInstancesBackup;
        }
        public void SetEffectInstancesBackup(EffectInstance[] effects)
        {
            m_axEffectInstancesBackup = effects;
        }
        public int[] Textures
        {
            get { return m_aiTextures; }
            set { m_aiTextures = value; }
        }
        public Mesh Mesh
        {
            get { return MeshData.Mesh; }
            set
            {
                MeshData xMeshData = MeshData;
                xMeshData.Mesh = value;
                MeshData = xMeshData;
            }
        }
        public Vector3 Centre
        {
            get { return m_xCentre; }
            set { m_xCentre = value; }
        }
        public float Radius
        {
            get { return m_fRadius; }
            set { m_fRadius = value; }
        }
        public Vector3 Scale
        {
            get { return m_xScale; }
            set { m_xScale = value; }
        }
    }

    class ModelExAllocateHierarchy : AllocateHierarchy
    {
        ModelExSystem m_xSystem;
        List<ModelExContainer> m_xContainers;
        public List<ModelExContainer> Containers { get { return m_xContainers; } }

        public ModelExAllocateHierarchy(ModelExSystem xSystem, List<ModelExContainer> xContainers)
            : base()
        {
            m_xSystem = xSystem;
            m_xContainers = xContainers;
        }

        public override Frame CreateFrame(string name)
        {
            return new ModelExFrame();
        }

        public override MeshContainer CreateMeshContainer(
            string name,
            MeshData meshData,
            ExtendedMaterial[] materials,
            EffectInstance[] effectInstances,
            GraphicsStream adjacency,
            SkinInformation skinInfo
        )
        {
            // Create the container
            ModelExContainer xContainer = new ModelExContainer();
            m_xContainers.Add(xContainer);

            // Set the container name
            xContainer.Name = name;

            // Set the mesh
            xContainer.MeshData = meshData;

            // Generate the adjacancy
            if (adjacency == null)
            {
                int[] aiAdjacency = new int[xContainer.Mesh.NumberFaces * 3];
                xContainer.Mesh.GenerateAdjacency(0.0f, aiAdjacency);
                xContainer.SetAdjacency(aiAdjacency);
            }
            else
            {
                xContainer.SetAdjacency(adjacency);
            }

            // Compute the normals
            if ((xContainer.Mesh.VertexFormat & VertexFormats.Normal) != VertexFormats.Normal)
            {
                xContainer.Mesh = xContainer.Mesh.Clone(
                    MeshFlags.Managed,
                    xContainer.Mesh.VertexFormat | VertexFormats.Normal,
                    xContainer.Mesh.Device
                );
                GraphicsStream xAdjacency = xContainer.GetAdjacencyStream();
                xContainer.Mesh.ComputeNormals(xAdjacency);
            }

            // Compute the bounding sphere
            GraphicsStream xVertexData = xContainer.Mesh.VertexBuffer.Lock(0, 0, LockFlags.None);
            Vector3 xCentre = Vector3.Empty;
            float fRadius = Geometry.ComputeBoundingSphere(
                xVertexData,
                xContainer.Mesh.NumberVertices,
                xContainer.Mesh.VertexFormat,
                out xCentre
            );
            xContainer.Mesh.VertexBuffer.Unlock();
            xContainer.Centre = xCentre;
            xContainer.Radius = fRadius;

            // Set the skin info
            if (skinInfo != null)
            {
                xContainer.SkinInformation = skinInfo;
            }

            // Create the materials
            if (materials == null || materials.Length == 0)
            {
                materials = new ExtendedMaterial[xContainer.Mesh.NumberAttributes];
            }
            for (int m = 0; m < materials.Length; m++)
            {
                if (materials[m].TextureFilename == null)
                {
                    materials[m].TextureFilename = "";
                }
            }

            int[] aiTextures = new int[xContainer.Mesh.NumberAttributes];

            xContainer.SetMaterials(materials);
            xContainer.Textures = aiTextures;

            // Create the effects
            if (effectInstances == null || effectInstances.Length == 0)
            {
                effectInstances = new EffectInstance[xContainer.Mesh.NumberAttributes];
            }
            for (int e = 0; e < effectInstances.Length; e++)
            {
                if (effectInstances[e].EffectFilename == null)
                {
                    effectInstances[e].EffectFilename = "";
                }
            }

            int[] aiEffects = new int[xContainer.Mesh.NumberAttributes];

            xContainer.SetEffectInstances(effectInstances);
            xContainer.SetEffectInstancesBackup(effectInstances);
            xContainer.Effects = aiEffects;

            return xContainer;
        }
    }

    class ModelExModelFile : ModelExFile
    {
        protected ModelExSystem m_xSystem;
        protected String m_strFileName;
        protected ModelFormat m_eFormat;
        protected ModelExFrame m_xFrame;
        protected List<ModelExContainer> m_xContainers;

        public String FileName { get { return m_strFileName; } }
        public ModelFormat Format { get { return m_eFormat; } }
        public ModelExFrame Frame { get { return m_xFrame; } }
        public ModelExContainer[] Containers { get { return m_xContainers.ToArray(); } }

        protected ModelExModelFile(ModelExSystem xSystem, String strFilePath)
        {
            m_xSystem = xSystem;
            m_strFileName = strFilePath;
            m_eFormat = ModelFormat.BinaryCompressed;
            m_xContainers = new List<ModelExContainer>();

            if (m_xSystem == null)
            {
                throw new Exception("Model file must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Model file must have a folder");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Model file must have a name");
            }

            Load(strFilePath);
        }

        protected ModelExModelFile(ModelExSystem xSystem,String strFilePath, ModelExFrame xFrame)
        {
            m_xSystem = xSystem;
            m_strFileName = strFilePath;
            m_eFormat = ModelFormat.BinaryCompressed;
            m_xContainers = new List<ModelExContainer>();

            if (m_xSystem == null)
            {
                throw new Exception("Texture file must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a parent or a folder");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a name");
            }

            Load(xFrame);
        }

        ~ModelExModelFile()
        {
            m_xSystem = null;
            m_strFileName = null;
        }

        public static ModelExModelFile Create(ModelExSystem xSystem, ModelExFile xParent, String strFilePath)
        {
            ModelExModelFile xModelFile = xSystem.FindModelFile(strFilePath);

            if (xModelFile == null)
            {
                xModelFile = new ModelExModelFile(xSystem, strFilePath);
                xSystem.AddModelFile(xModelFile);
            }

            return xModelFile;
        }

        public static ModelExModelFile Create(ModelExSystem xSystem, String strFilePath, ModelExFrame xFrame)
        {
            ModelExModelFile xModelFile = xSystem.FindModelFile(strFilePath);

            if (xModelFile == null)
            {
                xModelFile = new ModelExModelFile(xSystem, strFilePath, xFrame);
                xSystem.AddModelFile(xModelFile);
            }

            return xModelFile;
        }

        public void Load(String strFilePath)
        {
            FileStream xFileStream = null;
            BinaryReader xReader = null;
            AnimationRootFrame xAnimationRootFrame;
            bool bIsXFile = true;

            try
            {
                xFileStream = new FileStream(strFilePath, FileMode.Open, FileAccess.Read);
                xReader = new BinaryReader(xFileStream);

                String strMagicNumber = new String(xReader.ReadChars(4));
                if (strMagicNumber != "xof ")
                {
                    bIsXFile = false;
                }

                xReader.BaseStream.Position = 0x08;
                String strFileFormat = new String(xReader.ReadChars(4));
                switch (strFileFormat)
                {
                    case "txt ":
                        m_eFormat = ModelFormat.Text;
                        break;
                    case "tzip":
                        m_eFormat = ModelFormat.TextCompressed;
                        break;
                    case "bin ":
                        m_eFormat = ModelFormat.Binary;
                        break;
                    case "bzip":
                        m_eFormat = ModelFormat.BinaryCompressed;
                        break;
                    default:
                        bIsXFile = false;
                        break;
                }

                xReader.BaseStream.Position = 0x00;
                m_xContainers.Clear();

                ModelExAllocateHierarchy xAllocateHeirarchy = new ModelExAllocateHierarchy(m_xSystem, m_xContainers);
                xAnimationRootFrame = Mesh.LoadHierarchy(
                    xFileStream,
                    MeshFlags.Managed,
                    m_xSystem.Device,
                    xAllocateHeirarchy,
                    null
                );

                m_xFrame = (ModelExFrame)xAnimationRootFrame.FrameHierarchy;

                if (xAnimationRootFrame.FrameHierarchy == null)
                {
                    bIsXFile = false;
                }
            }
            catch
            {
                bIsXFile = false;
            }

            if (xReader != null)
            {
                xReader.Close();
            }

            if (xFileStream != null)
            {
                xFileStream.Close();
            }

            if (!bIsXFile)
            {
                throw new Exception("Not an X file");
            }

            if (m_xFrame.MeshContainer == null)
            {
                m_xFrame = (ModelExFrame)m_xFrame.FrameFirstChild;

                if (m_xFrame.MeshContainer == null)
                {
                    m_xFrame = (ModelExFrame)m_xFrame.FrameSibling;
                }
            }
        }

        public void Load(ModelExFrame xFrame)
        {
            m_eFormat = ModelFormat.BinaryCompressed;
            m_xFrame = xFrame;
            m_xContainers.Clear();
            m_xContainers.Add(xFrame.ModelExContainer);
        }

        public void Save(String strFilePath, ModelFormat eFormat)
        {
            if (eFormat == ModelFormat.LastUsed)
            {
                eFormat = m_eFormat;
            }

            D3DX.XFileFormat eXFileFormat = (D3DX.XFileFormat)eFormat;

            Mesh.SaveHierarchyToFile(
                strFilePath,
                eXFileFormat,
                m_xFrame,
                null,
                null
            );
        }
    }

    public abstract class ModelExModel
    {
        public abstract String Name { get; }
        public abstract String FileName { get; }

        public abstract void Save(String strFilePath, ModelFormat eFormat);
    }

    class ModelExModelList : ModelExList<ModelExModel>
    {
        List<ModelExModelInstance> m_xList;

        public List<ModelExModelInstance> List { get { return m_xList; } }
        public int Count { get { return m_xList.Count; } }
        public ModelExModel this[int m] { get { return m_xList[m]; } }

        public ModelExModelList()
        {
            m_xList = new List<ModelExModelInstance>();
        }

        public IEnumerator<ModelExModel> GetEnumerator()
        {
            for (int m = 0; m < m_xList.Count; m++)
            {
                yield return m_xList[m];
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }

    class ModelExModelInstance : ModelExModel
    {
        protected ModelExSystem m_xSystem;
        protected ModelExModelFile m_xFile;
        protected String m_strModelName;
        public override String Name { get { return m_strModelName; } }
        public override String FileName { get { return (m_xFile != null) ? m_xFile.FileName : null; } }
        public ModelExModelFile File { get { return m_xFile; } }
        public ModelExFrame Frame { get { return (m_xFile != null) ? m_xFile.Frame : null; } }

        public static implicit operator ModelExFrame(ModelExModelInstance xInstance)
        {
            return xInstance.Frame;
        }

        public ModelExModelInstance(ModelExSystem xSystem, String strFileName)
        {
            m_xSystem = xSystem;

            try
            {
                m_strModelName = Path.GetFileName(strFileName);
            }
            catch { }

            if (xSystem == null)
            {
                throw new Exception("Model instance must have a system");
            }

            if (m_strModelName == null)
            {
                throw new Exception("Model instance must have a file name");
            }
        }

        public void Load(String strFilePath)
        {
            try
            {
                m_xFile = ModelExModelFile.Create(m_xSystem, null, strFilePath);
            }
            catch
            {
                m_xFile = null;
            }
        }

        public virtual void Load(ModelExModelFile xModelFile)
        {
            m_xFile = xModelFile;
        }

        public override void Save(String strFilePath, ModelFormat eFormat)
        {
            if (m_xFile != null)
            {
                m_xFile.Save(strFilePath, eFormat);
            }
        }
    }
}
